#include <stdio.h>
#include "mpi.h"

#define MAX_PROCS 32
#define MAX_MEM   40960

int p_imin[MAX_PROCS], p_imax[MAX_PROCS];
int p_jmin[MAX_PROCS], p_jmax[MAX_PROCS];

int m_start, m_end;
int n_start, n_end;

int Y[MAX_MEM];
int TAG = 99;
MPI_Status status;

void main(int argc, char **argv)
{
   int M, N, P;
   int pm, pn;
   int m, n, MM, NN;
   int mdiv=1, ndiv=1;
   int count = 0;
   long  LONG;
   int i, j, k;
   int me;
   int comm[4];

   int imin[MAX_PROCS], imax[MAX_PROCS];
   int jmin[MAX_PROCS], jmax[MAX_PROCS];
 
   int fd;

   if (argc != 5) {
	printf("Usage: main columns rows processors memory\n");
	exit(0);
   }
   sscanf(argv[1], "%d", &M);
   sscanf(argv[2], "%d", &N);
   sscanf(argv[3], "%d", &P);
   sscanf(argv[4], "%ld", &LONG);

   MPI_Init( &argc, &argv) ;
   MPI_Comm_rank(MPI_COMM_WORLD, &me) ;

 if (me == 0) {
/*   printf("Input the Matrix size (M x N) : ");
   scanf("%d%d",&M,&N);
   printf("How many processors: ");
   scanf("%d",&P);
   printf("Maximum memory of each processor : ");
   scanf("%ld",&LONG);
*/

/* initializing the matrix */
   initialize();

/* Mesh the processors */
   process_mesh(P, &pm, &pn);

   MM = (M % pm) == 0 ? M / pm : M / pm + 1 ;
   NN = (N % pn) == 0 ? N / pn : N / pn + 1 ;

   matrix_part(M, N, MM, NN, pm, pn);
   print_mesh(pm, pn);

   partition(MM, NN, LONG, &m, &n, &mdiv, &ndiv);
   printf("m = %d n = %d mdiv = %d ndiv = %d\n", m, n, mdiv, ndiv);

/*
   for (i = 0; i< P-1; i++) {
      MPI_Send(&mdiv, 1, MPI_INT, i+1, TAG, MPI_COMM_WORLD) ;
      MPI_Send(&ndiv, 1, MPI_INT, i+1, TAG+1, MPI_COMM_WORLD) ;
   }
*/

/* Changed to broadcast */
/*   MPI_Bcast(&mdiv, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
   MPI_Bcast(&ndiv, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
*/
   comm[0] = mdiv;
   comm[1] = ndiv;
   MPI_Bcast(comm, 2, MPI_INT, 0, MPI_COMM_WORLD) ;

   if ((fd = creat("matrix.rst", 0666)) < 0) {
	printf("Sorry: result file matrix.rst couldn't be created.\n");
	exit(-1);
    }

    write(fd, &M, sizeof(int));
    write(fd, &N, sizeof(int));

   for (i = 0; i < mdiv; i++)
     for (j = 0; j < ndiv; j++) {
	for (k = 0; k < P; k++) {
	    imin[k] = p_imin[k] + i * m ;
/*	    imax[k] = i < mdiv-1 ? imin[k] + m - 1 : p_imax[k] ; */
	    imax[k] = imin[k]+m-1<=p_imax[k] ? imin[k] + m - 1 : p_imax[k] ; 
	    jmin[k] = p_jmin[k] + j * n ;
/*	    jmax[k] = j < ndiv-1 ? jmin[k] + n - 1 : p_jmax[k] ; */
	    jmax[k] = jmin[k]+n-1<=p_jmax[k] ? jmin[k] + n - 1 : p_jmax[k] ; 

/*  printf("processor %d : (%d %d) ---> (%d %d)\t", k, imin[k], jmin[k], 
		imax[k], jmax[k]); */

	    if (k < P-1) send_data(k, comm, imin[k], jmin[k], imax[k], jmax[k]);
	    else {
		m_start = imin[k]; 
		n_start = jmin[k];
		m_end = imax[k];
		n_end = jmax[k];
	    }
	}
/* printf("\n\n"); */

	compute(me);
	write_to_file(fd, m_start, n_start, m_end, n_end);
	for (k = 0; k < P-1; k++) {
	    if (imin[k] <= p_imax[k] && jmin[k] <= p_jmax[k]) { 
	        receive_result(k, imin[k], jmin[k], imax[k], jmax[k]);
	        write_to_file(fd, imin[k], jmin[k], imax[k], jmax[k]);
	    } 
	}
     }
   close(fd);
 }
 else {
/*
   	MPI_Recv(&mdiv, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
   	MPI_Recv(&ndiv, 1, MPI_INT, 0, TAG+1, MPI_COMM_WORLD, &status);
*/

/* Changed to broadcast */
/*   	MPI_Bcast(&mdiv, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
	MPI_Bcast(&ndiv, 1, MPI_INT, 0, MPI_COMM_WORLD) ;
*/
   	MPI_Bcast(comm, 2, MPI_INT, 0, MPI_COMM_WORLD) ;
	mdiv = comm[0];
	ndiv = comm[1];

	for (i = 0; i < mdiv; i++)
	    for (j = 0; j < ndiv; j++) {
		receive_data(comm);
		compute(me);
		send_result();
	    }
 }

   MPI_Finalize();
}

matrix_part(int M, int N, int MM, int NN, int pm, int pn)
{
   int i,j;
   int procs = 0;

   for (i = 0; i < pm; i++)
      for (j=0; j < pn; j++) {
	p_imin[procs] = i * MM;
	p_imax[procs] = i < pm-1 ? p_imin[procs] + MM -1 : M-1;
	p_jmin[procs] = j * NN;
	p_jmax[procs] = j < pn-1 ? p_jmin[procs] + NN -1 : N-1;
	procs ++;
      }
}

process_mesh(int P, int *m, int *n)
{
   int i=1;

   *m = 1; 
   *n = P;
   while (1) {
	if ( ++i > *n) break;
	if (!(P % i)) {
		*m = i;
		*n = P / i;
	}
   }
}

initialize()
{
}

partition(int M, int N, long LONG, int *m, int *n, int *mdiv, int *ndiv)
{
   *m = M;
   *n = N;
   while (*m * *n > LONG) {
	if ( *n > *m ) {
	   (*ndiv)++;
	   *n = N % *ndiv == 0 ? N / *ndiv : N / *ndiv + 1 ;
	}
	else {
	   (*mdiv)++;
	   *m = M % *mdiv == 0 ? M / *mdiv : M / *mdiv + 1 ;
	}
   }
}

send_data(int k, int *comm, int i_start, int j_start, int i_end, int j_end)
{
/*
   MPI_Send(&i_start, 1, MPI_INT, k+1, TAG, MPI_COMM_WORLD) ;
   MPI_Send(&j_start, 1, MPI_INT, k+1, TAG+1, MPI_COMM_WORLD) ;
   MPI_Send(&i_end, 1, MPI_INT, k+1, TAG+2, MPI_COMM_WORLD) ;
   MPI_Send(&j_end, 1, MPI_INT, k+1, TAG+3, MPI_COMM_WORLD) ;
*/

   *comm = i_start;
   *(comm+1) = j_start;
   *(comm+2) = i_end;
   *(comm+3) = j_end;
   MPI_Send(comm, 4, MPI_INT, k+1, TAG, MPI_COMM_WORLD) ;
}

receive_data(int *comm)
{
/*
   MPI_Recv(&m_start, 1, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
   MPI_Recv(&n_start, 1, MPI_INT, 0, TAG+1, MPI_COMM_WORLD, &status);
   MPI_Recv(&m_end, 1, MPI_INT, 0, TAG+2, MPI_COMM_WORLD, &status);
   MPI_Recv(&n_end, 1, MPI_INT, 0, TAG+3, MPI_COMM_WORLD, &status);
*/

   MPI_Recv(comm, 4, MPI_INT, 0, TAG, MPI_COMM_WORLD, &status);
   m_start = *comm;
   n_start = *(comm+1);
   m_end = *(comm+2);
   n_end = *(comm+3);
}

compute(int me)
{
   int i, j;
   int addr=0;

/*    printf("In processor %d : (%d %d) ====> (%d %d)\n", 
	me, m_start, n_start, m_end, n_end); */

   for (i = m_start; i <= m_end; i++) 
      for (j = n_start; j <= n_end; j++)  
	 Y[addr++] = i + j ;
}

receive_result(int k, int i_start, int j_start, int i_end, int j_end)
{
   MPI_Recv(Y, (i_end-i_start+1)*(j_end-j_start+1), MPI_INT, k+1, 
	TAG, MPI_COMM_WORLD, &status);
}

send_result()
{
   if (m_start <= m_end && n_start <= n_end) 
   MPI_Send(Y, (m_end-m_start+1)*(n_end-n_start+1), MPI_INT, 0, 
		TAG, MPI_COMM_WORLD) ;
}

print_mesh(int pm, int pn)
{
   int i, j;
   int procs = 0;

   printf("pm = %d\tpn = %d\n",pm, pn);
   for (i=0; i<pm; i++)
      for (j=0; j<pn; j++) {
	printf("(%d %d)\t(%d %d)\n",p_imin[procs], p_jmin[procs], 
		p_imax[procs], p_jmax[procs]);
	procs ++;
	}
}

write_to_file(int fd, int i_start, int j_start, int i_end, int j_end)
{
   int addr = 0;
   int i, j;

/*   printf(" Block (%d %d) =====> (%d %d)\n", i_start, j_start, i_end, j_end); */

   write(fd, &i_start, sizeof(int));
   write(fd, &j_start, sizeof(int));
   write(fd, &i_end, sizeof(int));
   write(fd, &j_end, sizeof(int));
   write(fd, Y, sizeof(int) * (i_end-i_start+1) * (j_end-j_start+1));

/*
   for (i = i_start; i <= i_end; i++) {
	for (j = j_start; j <= j_end; j++) 
	   printf("%d ",Y[addr++]);
	printf("\n"); 
   }
   printf("\n\n\n");
*/
}
